@@ -1,17 +1,22 @@
 # coding: utf-8
+import platform
+import sys
 from datetime import date, datetime, timedelta, tzinfo
-from functools import partial
+from functools import partial, wraps
+from types import ModuleType
 
 import pytest
 import pytz
 import six
-import sys
 
-from apscheduler.util import (
-    asint, asbool, astimezone, convert_to_datetime, datetime_to_utc_timestamp,
-    utc_timestamp_to_datetime, timedelta_seconds, datetime_ceil, get_callable_name, obj_to_ref,
-    ref_to_obj, maybe_ref, check_callable_args, datetime_repr, repr_escape)
-from tests.conftest import minpython, maxpython
+from apscheduler.job import Job
+from apscheduler.util import (asbool, asint, astimezone, check_callable_args,
+                              convert_to_datetime, datetime_ceil,
+                              datetime_repr, datetime_to_utc_timestamp,
+                              get_callable_name, maybe_ref, obj_to_ref,
+                              ref_to_obj, repr_escape, timedelta_seconds,
+                              utc_timestamp_to_datetime)
+from tests.conftest import maxpython, minpython
 
 try:
     from unittest.mock import Mock
@@ -40,6 +45,10 @@ class DummyClass(object):
             pass
 
 
+class InheritedDummyClass(Job):
+    pass
+
+
 class TestAsint(object):
     @pytest.mark.parametrize('value', ['5s', 'shplse'], ids=['digit first', 'text'])
     def test_invalid_value(self, value):
@@ -89,7 +98,7 @@ class TestAstimezone(object):
         assert 'Only timezones from the pytz library are supported' in str(exc.value)
 
     def test_bad_local_timezone(self):
-        zone = Mock(tzinfo, localize=None, normalize=None, zone='local')
+        zone = Mock(tzinfo, localize=None, normalize=None, tzname=lambda dt: 'local')
         exc = pytest.raises(ValueError, astimezone, zone)
         assert 'Unable to determine the name of the local timezone' in str(exc.value)
 
@@ -105,9 +114,15 @@ class TestConvertToDatetime(object):
         (datetime(2009, 8, 1, 5, 6, 12), datetime(2009, 8, 1, 5, 6, 12)),
         ('2009-8-1', datetime(2009, 8, 1)),
         ('2009-8-1 5:16:12', datetime(2009, 8, 1, 5, 16, 12)),
+        ('2009-8-1T5:16:12Z', datetime(2009, 8, 1, 5, 16, 12, tzinfo=pytz.utc)),
+        ('2009-8-1T5:16:12+02:30',
+         pytz.FixedOffset(150).localize(datetime(2009, 8, 1, 5, 16, 12))),
+        ('2009-8-1T5:16:12-05:30',
+         pytz.FixedOffset(-330).localize(datetime(2009, 8, 1, 5, 16, 12))),
         (pytz.FixedOffset(-60).localize(datetime(2009, 8, 1)),
          pytz.FixedOffset(-60).localize(datetime(2009, 8, 1)))
-    ], ids=['None', 'date', 'datetime', 'date as text', 'datetime as text', 'existing tzinfo'])
+    ], ids=['None', 'date', 'datetime', 'date as text', 'datetime as text', 'utc', 'tzoffset',
+            'negtzoffset', 'existing tzinfo'])
     def test_date(self, timezone, input, expected):
         returned = convert_to_datetime(input, timezone, None)
         if expected is not None:
@@ -133,11 +148,6 @@ class TestConvertToDatetime(object):
         returned = convert_to_datetime('2009-8-1', 'UTC', None)
         assert returned == datetime(2009, 8, 1, tzinfo=pytz.utc)
 
-    def test_bad_timezone(self):
-        exc = pytest.raises(TypeError, convert_to_datetime, '2009-8-1', tzinfo(), None)
-        assert str(exc.value) == ('Only pytz timezones are supported (need the localize() and '
-                                  'normalize() methods)')
-
 
 def test_datetime_to_utc_timestamp(timezone):
     dt = timezone.localize(datetime(2014, 3, 12, 5, 40, 13, 254012))
@@ -189,28 +199,41 @@ class TestGetCallableName(object):
 
 
 class TestObjToRef(object):
-    @pytest.mark.parametrize('input', [partial(DummyClass.meth)], ids=['partial/bound method'])
-    def test_no_ref_found(self, input):
-        exc = pytest.raises(ValueError, obj_to_ref, input)
-        assert 'Cannot determine the reference to ' in str(exc.value)
+    @pytest.mark.parametrize('obj, error', [
+        (partial(DummyClass.meth), 'Cannot create a reference to a partial()'),
+        (lambda: None, 'Cannot create a reference to a lambda')
+    ], ids=['partial', 'lambda'])
+    def test_errors(self, obj, error):
+        exc = pytest.raises(ValueError, obj_to_ref, obj)
+        assert str(exc.value) == error
+
+    @pytest.mark.skipif(sys.version_info[:2] < (3, 3),
+                        reason='Requires __qualname__ (Python 3.3+)')
+    def test_nested_function_error(self):
+        def nested():
+            pass
+
+        exc = pytest.raises(ValueError, obj_to_ref, nested)
+        assert str(exc.value) == 'Cannot create a reference to a nested function'
 
     @pytest.mark.parametrize('input,expected', [
-        pytest.mark.skipif(sys.version_info[:2] == (3, 2),
-                           reason="Unbound methods can't be resolved on Python 3.2")(
-            (DummyClass.meth, 'tests.test_util:DummyClass.meth')
-        ),
+        (DummyClass.meth, 'tests.test_util:DummyClass.meth'),
         (DummyClass.classmeth, 'tests.test_util:DummyClass.classmeth'),
-        pytest.mark.skipif(sys.version_info < (3, 3),
-                           reason="Requires __qualname__ (Python 3.3+)")(
-            (DummyClass.InnerDummyClass.innerclassmeth,
-             'tests.test_util:DummyClass.InnerDummyClass.innerclassmeth')
+        pytest.param(
+            DummyClass.InnerDummyClass.innerclassmeth,
+            'tests.test_util:DummyClass.InnerDummyClass.innerclassmeth',
+            marks=[pytest.mark.skipif(sys.version_info < (3, 3),
+                                      reason="Requires __qualname__ (Python 3.3+)")]
         ),
-        pytest.mark.skipif(sys.version_info < (3, 3),
-                           reason="Requires __qualname__ (Python 3.3+)")(
-            (DummyClass.staticmeth, 'tests.test_util:DummyClass.staticmeth')
+        pytest.param(
+            DummyClass.staticmeth,
+            'tests.test_util:DummyClass.staticmeth',
+            marks=[pytest.mark.skipif(sys.version_info < (3, 3),
+                                      reason="Requires __qualname__ (Python 3.3+)")]
         ),
         (timedelta, 'datetime:timedelta'),
-    ], ids=['unbound method', 'class method', 'inner class method', 'static method', 'timedelta'])
+    ], ids=['class method', 'inner class method', 'static method', 'inherited class method',
+            'timedelta'])
     def test_valid_refs(self, input, expected):
         assert obj_to_ref(input) == expected
 
@@ -220,6 +243,15 @@ class TestRefToObj(object):
         from logging.handlers import RotatingFileHandler
         assert ref_to_obj('logging.handlers:RotatingFileHandler') is RotatingFileHandler
 
+    def test_complex_path(self):
+        pkg1 = ModuleType('pkg1')
+        pkg1.pkg2 = 'blah'
+        pkg2 = ModuleType('pkg1.pkg2')
+        pkg2.varname = 'test'
+        sys.modules['pkg1'] = pkg1
+        sys.modules['pkg1.pkg2'] = pkg2
+        assert ref_to_obj('pkg1.pkg2:varname') == 'test'
+
     @pytest.mark.parametrize('input,error', [
         (object(), TypeError),
         ('module', ValueError),
@@ -292,6 +324,8 @@ class TestCheckCallableArgs(object):
         check_callable_args(object().__setattr__, ('blah', 1), {})
 
     @minpython(3, 4)
+    @pytest.mark.skipif(platform.python_implementation() == 'PyPy',
+                        reason='PyPy does not expose signatures of builtins')
     def test_positional_only_args(self):
         """
         Tests that an attempt to use keyword arguments for positional-only arguments raises an
@@ -314,3 +348,17 @@ class TestCheckCallableArgs(object):
         exc = pytest.raises(ValueError, check_callable_args, func, [1], {})
         assert str(exc.value) == ('The following keyword-only arguments have not been supplied in '
                                   'kwargs: y')
+
+    def test_wrapped_func(self):
+        """
+        Test that a wrapped function can be scheduled even if it cannot accept the arguments given
+        in add_job() if the wrapper can.
+        """
+        def func():
+            pass
+
+        @wraps(func)
+        def wrapper(arg):
+            func()
+
+        check_callable_args(wrapper, (1,), {})
